import warnings

import zope.component.zcml
import zope.interface
import zope.schema
from zope.component.security import PublicPermission
from zope.configuration.name import resolve

from .interfaces import IEngineFactory
from .interfaces import IScopedSession
from .utility import EngineFactory


class IEngineDirective(zope.interface.Interface):
    """Registers a database engine factory."""

    url = zope.schema.URI(
        title='Database URL',
        description="e.g. 'sqlite:///:memory:'.",
        required=True)

    name = zope.schema.Text(
        title="Engine name",
        description="Empty if this is the default engine.",
        required=False,
        default='')

    convert_unicode = zope.schema.Bool(
        title='Convert all string columns to unicode',
        description='This setting makes the SQLAlchemy String column type '
                    'equivalent to UnicodeString. Do not use this unless '
                    'there is a good reason not to use standard '
                    'UnicodeString columns',
        required=False,
        default=False)

    echo = zope.schema.Bool(
        title='Echo SQL statements',
        description='Enable logging statements for debugging.',
        required=False,
        default=None)

    setup = zope.schema.BytesLine(
        title='After engine creation hook',
        description='Callback for creating mappers etc. '
                    'One argument is passed, the engine',
        required=False,
        default=None)

    # Connection pooling options - probably only works on SQLAlchemy 0.5 and up

    pool_size = zope.schema.Int(
        title="The size of the pool to be maintained",
        description="Defaults to 5 in SQLAlchemy.",
        required=False)

    max_overflow = zope.schema.Int(
        title="The maximum overflow size of the pool.",
        description="When the number of checked-out connections "
                    "reaches the size set in pool_size, additional "
                    "connections will be returned up to this limit. "
                    "Defaults to 10 in SQLAlchemy",
        required=False)

    pool_recycle = zope.schema.Int(
        title="Number of seconds between connection recycling",
        description="Upon checkout, if this timeout is "
                    "surpassed the connection will be closed and "
                    "replaced with a newly opened connection",
        required=False)

    pool_timeout = zope.schema.Int(
        title="The number of seconds to wait before giving up on "
              "returning a connection.",
        description="Defaults to 30 in SQLAlchemy if not set",
        required=False)


class ISessionDirective(zope.interface.Interface):
    """Registers a database scoped session"""

    name = zope.schema.Text(
        title="Scoped session name",
        description="Empty if this is the default session.",
        required=False,
        default="")

    twophase = zope.schema.Bool(
        title='Use two-phase commit',
        description='Session should use two-phase commit',
        required=False,
        default=False)

    engine = zope.schema.Text(
        title="Engine name",
        description="Empty if this is to use the default engine.",
        required=False,
        default="")

    factory = zope.schema.DottedName(
        title='Scoped Session utility factory',
        description='GloballyScopedSession by default',
        required=False,
        default="z3c.saconfig.utility.GloballyScopedSession")


def engine(_context, url, name="", convert_unicode=False,
           echo=None, setup=None, twophase=False,
           pool_size=None, max_overflow=None, pool_recycle=None,
           pool_timeout=None):

    if convert_unicode:  # pragma: no cover
        warnings.warn(
            '`convert_unicode` is no longer suported by SQLAlchemy, so it is'
            ' ignored here.', DeprecationWarning)

    kwargs = {
        'echo': echo,
    }

    # Only add these if they're actually set, since we want to let SQLAlchemy
    # control the defaults
    if pool_size is not None:
        kwargs['pool_size'] = pool_size
    if max_overflow is not None:
        kwargs['max_overflow'] = max_overflow
    if pool_recycle is not None:
        kwargs['pool_recycle'] = pool_recycle
    if pool_timeout is not None:
        kwargs['pool_timeout'] = pool_timeout

    factory = EngineFactory(url, **kwargs)

    zope.component.zcml.utility(
        _context,
        provides=IEngineFactory,
        component=factory,
        permission=PublicPermission,
        name=name)

    if setup:
        if isinstance(setup, bytes):
            setup = setup.decode()

        if _context.package is None:
            callback = resolve(setup)
        else:
            callback = resolve(setup, package=_context.package.__name__)
        _context.action(
            discriminator=(IEngineFactory, name),
            callable=callback,
            args=(factory(), ),
            order=9999)


def session(_context, name="", engine="", twophase=False,
            factory="z3c.saconfig.utility.GloballyScopedSession"):
    if _context.package is None:
        ScopedSession = resolve(factory)
    else:
        ScopedSession = resolve(factory, package=_context.package.__name__)
    scoped_session = ScopedSession(engine=engine, twophase=twophase)

    zope.component.zcml.utility(
        _context,
        provides=IScopedSession,
        component=scoped_session,
        permission=PublicPermission,
        name=name)



from sqlalchemy.orm import scoped_session
from zope import component

from z3c.saconfig.interfaces import IScopedSession


def session_factory(name=''):
    """This is used by scoped session to create a new Session object.

    It delegates to a IScopedSession utility.
    """
    utility = component.getUtility(IScopedSession, name)
    return utility.sessionFactory()


def scopefunc(name=''):
    """This is used by scoped session to distinguish between sessions.

    It delegates to a IScopedSession utility.
    """
    utility = component.getUtility(IScopedSession, name)
    return utility.scopeFunc()

# this is framework central configuration. Use a IScopedSession utility
# to define behavior.


Session = scoped_session(session_factory, scopefunc)

_named_scoped_sessions = {'': Session}


def named_scoped_session(name):
    try:
        return _named_scoped_sessions[name]
    except KeyError:
        return _named_scoped_sessions.setdefault(
            name,
            scoped_session(
                lambda: session_factory(name),
                lambda: scopefunc(name)))



from z3c.saconfig.scopedsession import Session
from z3c.saconfig.scopedsession import named_scoped_session
from z3c.saconfig.utility import EngineFactory
from z3c.saconfig.utility import GloballyScopedSession
from z3c.saconfig.utility import SiteScopedSession


__all__ = [
    'Session',
    'named_scoped_session',
    'GloballyScopedSession',
    'SiteScopedSession',
    'EngineFactory',
]



from zope.interface import Attribute
from zope.interface import Interface
from zope.interface import implementer


class IScopedSession(Interface):
    """A utility that plugs into SQLAlchemy's scoped session machinery.

    The idea is that you'd either register a IScopedSession utility globally,
    for simple configurations, or locally, if you want to have the ability
    to transparently use a different engine and session configuration per
    database.
    """
    def sessionFactory():
        """Create a SQLAlchemy session.

        Typically you'd use sqlalchemy.orm.create_session to create
        the session here.
        """

    def scopeFunc():
        """Determine the scope of the session.

        This can be used to scope the session per thread, per Zope 3 site,
        or otherwise. Return an immutable value to scope the session,
        like a thread id, or a tuple with thread id and application id.
        """


class ISiteScopedSession(IScopedSession):
    """A utility that makes sessions be scoped by site.
    """
    def siteScopeFunc():
        """Returns a unique id per site.
        """


class IEngineFactory(Interface):
    """A utility that maintains an SQLAlchemy engine.

    If the engine isn't created yet, it will create it. Otherwise the
    engine will be cached.
    """

    def __call__():
        """Get the engine.

        This creates the engine if this factory was not used before,
        otherwise returns a cached version.
        """

    def configuration():
        """Returns the engine configuration in the form of an args, kw tuple.

        Return the parameters used to create an engine as a tuple with
        an args list and a kw dictionary.
        """

    def reset():
        """Reset the cached engine (if any).

        This causes the engine to be recreated on next use.
        """


class IEngineCreatedEvent(Interface):
    """An SQLAlchemy engine has been created.

    Hook into this event to do setup that can only be performed with
    an active engine.
    """
    engine = Attribute("The engine that was just created.")

    engine_args = Attribute("List of arguments given to SQLAlchemy "
                            "create_engine")

    engine_kw = Attribute("Dictionary of keyword attributes given to "
                          "SQLAlchemy create_engine")


@implementer(IEngineCreatedEvent)
class EngineCreatedEvent:

    def __init__(self, engine, engine_args, engine_kw):
        self.engine = engine
        self.engine_args = engine_args
        self.engine_kw = engine_kw



"""
Some reusable, standard implementations of IScopedSession.
"""
import threading
import time
from threading import get_ident

import sqlalchemy
from zope import component
from zope.event import notify
from zope.interface import implementer
from zope.sqlalchemy import register

from z3c.saconfig.interfaces import EngineCreatedEvent
from z3c.saconfig.interfaces import IEngineFactory
from z3c.saconfig.interfaces import IScopedSession
from z3c.saconfig.interfaces import ISiteScopedSession


SESSION_DEFAULTS = dict(
    autocommit=False,
    autoflush=True,
)


@implementer(IScopedSession)
class GloballyScopedSession:
    """A globally scoped session.

    Register this as a global utility to have just one kind of session
    per Zope instance. All applications in this instance will share the
    same session.

    To register as a global utility you may need to register it with
    a custom factory, or alternatively subclass it and override __init__
    to pass the right arguments to the superclasses __init__.
    """

    def __init__(self, engine='', **kw):
        """Pass keywords arguments for sqlalchemy.orm.create_session.

        The `engine` argument is the name of a utility implementing
        IEngineFactory.

        Note that GloballyScopedSesssion does have different defaults than
        ``create_session`` for various parameters where it makes sense
        for Zope integration, namely:

        autocommit = False
        autoflush = True
        extension = ZopeTransactionExtension()

        Normally you wouldn't pass these in, but if you have the need
        to override them, you could.
        """
        self.engine = engine
        self.kw = _zope_session_defaults(kw)

    def sessionFactory(self):
        kw = self.kw.copy()
        if 'bind' not in kw:
            engine_factory = component.getUtility(IEngineFactory,
                                                  name=self.engine)
            kw['bind'] = engine_factory()
        session = sqlalchemy.orm.create_session(**kw)
        register(session)
        return session

    def scopeFunc(self):
        return get_ident()


def _zope_session_defaults(kw):
    """Adjust keyword parameters with proper defaults for Zope.
    """

    d = SESSION_DEFAULTS.copy()
    d.update(kw)

    return d


@implementer(ISiteScopedSession)
class SiteScopedSession:
    """A session that is scoped per site.

    Even though this makes the sessions scoped per site,
    the utility can be registered globally to make this work.

    Creation arguments as for GloballyScopedSession, except that no ``bind``
    parameter should be passed. This means it is possible to create
    a SiteScopedSession utility without passing parameters to its constructor.
    """

    def __init__(self, engine='', **kw):
        assert 'bind' not in kw
        self.engine = engine
        self.kw = _zope_session_defaults(kw)

    def sessionFactory(self):
        engine_factory = component.getUtility(IEngineFactory,
                                              name=self.engine)
        kw = self.kw.copy()
        kw['bind'] = engine_factory()
        session = sqlalchemy.orm.create_session(**kw)
        register(session)
        return session

    def scopeFunc(self):
        return (get_ident(), self.siteScopeFunc())

    def siteScopeFunc(self):
        raise NotImplementedError

# Credits: This method of storing engines lifted from zope.app.cache.ram


_COUNTER = 0
_COUNTER_LOCK = threading.Lock()

_ENGINES = {}
_ENGINES_LOCK = threading.Lock()


@implementer(IEngineFactory)
class EngineFactory:
    """An engine factory.

    If you need engine connection parameters to be different per site,
    EngineFactory should be registered as a local utility in that
    site.

    If you want this utility to be persistent, you should subclass it
    and mixin Persistent. You could then manage the parameters
    differently than is done in this __init__, for instance as
    attributes, which is nicer if you are using Persistent (or Zope 3
    schema). In this case you need to override the configuration method.
    """

    def __init__(self, *args, **kw):
        self._args = args
        self._kw = kw
        self._key = self._getKey()

    def _getKey(self):
        """Get a unique key"""
        global _COUNTER
        _COUNTER_LOCK.acquire()
        try:
            _COUNTER += 1
            return "%s_%f_%d" % (id(self), time.time(), _COUNTER)
        finally:
            _COUNTER_LOCK.release()

    def __call__(self):
        # optimistically try get without lock
        engine = _ENGINES.get(self._key, None)
        if engine is not None:
            return engine
        # no engine, lock and redo
        _ENGINES_LOCK.acquire()
        try:
            # need to check, another thread may have got there first
            if self._key not in _ENGINES:
                args, kw = self.configuration()
                _ENGINES[self._key] = engine = sqlalchemy.create_engine(
                    *args, **kw)
                notify(EngineCreatedEvent(engine, args, kw))
            return _ENGINES[self._key]
        finally:
            _ENGINES_LOCK.release()

    def configuration(self):
        """Returns engine parameters.

        This can be overridden in a subclass to retrieve the parameters
        from some other place.
        """
        return self._args, self._kw

    def reset(self):
        _ENGINES_LOCK.acquire()
        try:
            if self._key not in _ENGINES:
                return
            # XXX is disposing the right thing to do?
            _ENGINES[self._key].dispose()
            del _ENGINES[self._key]
        finally:
            _ENGINES_LOCK.release()



# You may want to run the tests with your database. To do so set the
# environment variable TEST_DSN to the connection url. e.g.:
#
# export TEST_DSN=postgres://plone:plone@localhost/test
# export TEST_DSN=mssql://plone:plone@/test?dsn=mydsn

# Since the test exercise what happens with two different DSNs
# locally, you need to also set up a different TEST_DSN2.

import doctest
import os
import unittest

import zope.component.eventtesting
# To test in twophase commit mode export TEST_TWOPHASE=True
#
# NOTE: The sqlite that ships with Mac OS X 10.4 is buggy.
# Install a newer version (3.5.6) and rebuild pysqlite2 against it.
from zope import component
from zope.interface import registry
from zope.testing import cleanup
from zope.testing.cleanup import addCleanUp


TEST_TWOPHASE = bool(os.environ.get('TEST_TWOPHASE'))
TEST_DSN = os.environ.get('TEST_DSN', 'sqlite:///:memory:')
TEST_DSN1 = TEST_DSN
# this can reuse TEST_DSN1 in the default case, as we can open another
# in-memory database. You can't do this for other databases however.
TEST_DSN2 = os.environ.get('TEST_DSN', TEST_DSN1)


def setUpReadMe(test):
    # set up special local component architecture
    setHooks()
    # set up event handling
    zope.component.eventtesting.setUp(test)


def tearDownReadMe(test):
    # clean up Zope
    cleanup.cleanUp()

    # clean up SQLAlchemy
    Base = test.globs['Base']
    engine = test.globs['engine']
    Base.metadata.drop_all(engine)

# a very simple implementation of setSite and getSite so we don't have
# to rely on zope.app.component just for our tests


_site = None


class DummySite:
    def __init__(self, id):
        self.id = id
        self._sm = SiteManager()

    def getSiteManager(self):
        return self._sm


class SiteManager(registry.Components):
    def __init__(self):
        super().__init__()
        self.__bases__ = (component.getGlobalSiteManager(),)


def setSite(site=None):
    global _site
    _site = site


def getSite():
    return _site


def adapter_hook(interface, object, name='', default=None):
    try:
        return getSiteManager().adapters.adapter_hook(
            interface, object, name, default)
    except component.interfaces.ComponentLookupError:
        return default


def getSiteManager(context=None):
    if _site is not None:
        return _site.getSiteManager()
    return component.getGlobalSiteManager()


def setHooks():
    component.adapter_hook.sethook(adapter_hook)
    component.getSiteManager.sethook(getSiteManager)


def resetHooks():
    component.adapter_hook.reset()
    component.getSiteManager.reset()

# make sure hooks get cleaned up after tests are run


addCleanUp(resetHooks)


def engine_subscriber(engine):
    print(f'got: {engine!s} ')


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    globs = {
        'DummySite': DummySite,
        'setSite': setSite,
        'getSite': getSite,
        'TEST_DSN': TEST_DSN,
        'TEST_DSN1': TEST_DSN1,
        'TEST_DSN2': TEST_DSN2,
        'TEST_TWOPHASE': TEST_TWOPHASE,
    }

    suite = unittest.TestSuite()

    suite.addTest(doctest.DocFileSuite(
        'README.rst',
        optionflags=optionflags,
        setUp=setUpReadMe,
        tearDown=tearDownReadMe,
        globs=globs))
    return suite



import os

from setuptools import setup


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


long_description = read('src/z3c/saconfig/README.rst') + \
    '\n' + read('CHANGES.rst')

setup(name='z3c.saconfig',
      version='2.0',
      description="Minimal SQLAlchemy ORM session configuration for Zope",
      long_description=long_description,
      # Get more strings from https://pypi.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'Framework :: Zope :: 3',
          'Framework :: Zope :: 4',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Zope Public License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
          'Programming Language :: Python :: 3.13',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Internet :: WWW/HTTP',
          'Topic :: Software Development :: Libraries',
          'Topic :: Database',
      ],
      keywords='zope relational sqlalchemy component integration',
      author='Martijn Faassen',
      author_email='faassen@startifact.com',
      url='https://github.com/zopefoundation/z3c.saconfig/',
      license='ZPL-2.1',
      include_package_data=True,
      zip_safe=False,
      python_requires='>=3.9',
      install_requires=[
          'setuptools',
          'zope.sqlalchemy>=0.5',
          'zope.interface',
          'zope.component',
          'zope.hookable',
          'zope.security',
          'zope.event',
          'zope.configuration',
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      extras_require=dict(
          test=['zope.testing'],
      ),
      )
